# Copyright (c) 2016-2021, Thomas Larsson
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import bpy
import os
from mathutils import *
from .error import *
from .utils import *

Converters = {}
TwistBones = {}
RestPoses = {}
Parents = {}
IkPoses = {}

#-------------------------------------------------------------
#   Save current pose
#-------------------------------------------------------------

class DAZ_OT_SaveCurrentPose(DazOperator, B.JsonExportFile, B.SkelPoseBool, IsArmature):
    bl_idname = "daz.save_current_pose"
    bl_label = "Save Current Pose"
    bl_options = {'UNDO'}

    def invoke(self, context, event):
        context.window_manager.fileselect_add(self)
        return {'RUNNING_MODAL'}

    def run(self, context):
        rig = context.object
        rolls = {}
        bpy.ops.object.mode_set(mode='EDIT')
        for eb in rig.data.edit_bones:
            rolls[eb.name] = eb.roll
        bpy.ops.object.mode_set(mode='OBJECT')

        from collections import OrderedDict
        struct = OrderedDict()
        struct["character"] = rig.name
        struct["scale"] = 1.0

        if self.skeleton:
            skel = {}
            struct["skeleton"] = skel
            for pb in rig.pose.bones:
                skel[pb.name]  = [pb.rotation_mode, rolls[pb.name]]

        if self.pose:
            pose = {}
            struct["pose"] = pose
            for pb in rig.pose.bones:
                if pb.bone.select:
                    euler = pb.matrix.to_3x3().to_euler()
                    pose[pb.name] = (list(euler), list(pb.bone.DazOrient), pb.DazRotMode)

        from .load_json import saveJson
        saveJson(struct, self.filepath)

#-------------------------------------------------------------
#   Load pose
#-------------------------------------------------------------

def getCharacter(rig):
    if rig.DazMesh:
        char = rig.DazMesh.lower().replace("-","_").replace("genesis", "genesis_")
        if char[-1] == "_":
            char = char[:-1]
        print("Character: %s" % char)
        return char
    else:
        return None


def loadRestPoseEntry(character, table, folder):
    import json
    from .fileutils import safeOpen
    if character in table.keys():
        return
    filepath = os.path.join(folder, character +  ".json")
    print("Load", filepath)
    if not os.path.exists(filepath):
        raise DazError("File %s    \n does not exist" % filepath)
    else:
        with safeOpen(filepath, "r") as fp:
            data = json.load(fp)
    table[character] = data


def getOrientation(character, bname, rig):
    from .globvars import theRestPoseFolder
    global RestPoses
    if rig and bname in rig.pose.bones.keys():
        pb = rig.pose.bones[bname]
        return pb.bone.DazOrient, pb.DazRotMode

    loadRestPoseEntry(character, RestPoses, theRestPoseFolder)
    bones = RestPoses[character]["pose"]
    if bname in bones.keys():
        vec, orient, xyz = bones[bname]
        return orient, xyz
    else:
        return None, "XYZ"


def getParentCharacter(character):
    from .globvars import theRestPoseFolder
    global RestPoses
    loadRestPoseEntry(character, RestPoses, theRestPoseFolder)
    if "parent" in RestPoses[character].keys():
        parent = RestPoses[character]["parent"]
        return parent.lower().replace(" ", "_")
    else:
        return character


def getParent(character, bname):
    from .globvars import theParentsFolder
    global Parents
    parent = getParentCharacter(character)
    loadRestPoseEntry(parent, Parents, theParentsFolder)
    parents = Parents[parent]["parents"]
    if bname in parents.keys() and parents[bname]:
        return parents[bname]
    else:
        return None


def loadPose(rig, character, table, modify):
    root = None
    for pb in rig.pose.bones:
        if pb.parent is None:
            root = pb
            break
    if "skeleton" in table[character]:
        modifySkeleton(rig, table[character]["skeleton"])
    loadBonePose(root, table[character]["pose"])


def modifySkeleton(rig, skel):
    bpy.ops.object.mode_set(mode='EDIT')
    for eb in rig.data.edit_bones:
        bname = getBoneName(eb.name, skel)
        if bname in skel.keys():
            eb.roll = skel[bname][1]
    bpy.ops.object.mode_set(mode='OBJECT')
    for pb in rig.pose.bones:
        bname = getBoneName(pb.name, skel)
        if bname in skel.keys():
            pb.rotation_mode = skel[bname][0]


def getBoneName(bname, bones):
    if bname in bones.keys():
        return bname
    elif (bname[-3:] == "Drv" and
          bname[:-3] in bones.keys()):
        return bname[:-3]
    elif (bname[-4:] == "Copy" and
          bname[:-4] in bones.keys()):
        return bname[:-4]
    else:
        return None


def loadBonePose(pb, pose):
    pbname = getBoneName(pb.name, pose)
    if pbname and pb.name[:-4] != "Copy":
        vec, pb.bone.DazOrient, pb.DazRotMode = pose[pbname]
        euler = Euler(vec)
        mat = euler.to_matrix()
        rmat = pb.bone.matrix_local.to_3x3()
        if pb.parent:
            par = pb.parent
            rmat = Mult2(par.bone.matrix_local.to_3x3().inverted(), rmat)
            mat = Mult2(par.matrix.to_3x3().inverted(), mat)
        bmat = Mult2(rmat.inverted(), mat)
        pb.matrix_basis = bmat.to_4x4()
        for n in range(3):
            if pb.lock_rotation[n]:
                pb.rotation_euler[n] = 0
        bpy.ops.object.mode_set(mode='EDIT')
        bpy.ops.object.mode_set(mode='OBJECT')

    if pb.name != "head":
        for child in pb.children:
            loadBonePose(child, pose)


class DAZ_OT_LoadPose(DazOperator, B.JsonFile, B.SingleFile, IsArmature):
    bl_idname = "daz.load_pose"
    bl_label = "Load Pose"
    bl_options = {'UNDO'}

    def run(self, context):
        folder = os.path.dirname(self.filepath)
        character = os.path.splitext(os.path.basename(self.filepath))[0]
        table = {}
        loadRestPoseEntry(character, table, folder)
        loadPose(context.object, character, table, False)
        print("Pose %s loaded" % self.filepath)

    def invoke(self, context, event):
        context.window_manager.fileselect_add(self)
        return {'RUNNING_MODAL'}

#-------------------------------------------------------------
#   Optimize pose for IK
#-------------------------------------------------------------

class DAZ_OT_OptimizePose(DazPropsOperator, B.MergeRigs, IsArmature):
    bl_idname = "daz.optimize_pose"
    bl_label = "Optimize Pose For IK"
    bl_description = "Optimize pose for IK.\nIncompatible with pose loading and body morphs."
    bl_options = {'UNDO'}

    def draw(self, context):
        self.layout.prop(self, "useApplyRestPose")


    def run(self, context):
        from .globvars import theIkPoseFolder
        from .merge import applyRestPoses
        rig = context.object
        char = getCharacter(rig)
        if char is None:
            raise DazError("Did not recognize character")
        loadRestPoseEntry(char, IkPoses, theIkPoseFolder)
        loadPose(rig, char, IkPoses, False)
        if self.useApplyRestPose:
            applyRestPoses(context, rig, [])

#-------------------------------------------------------------
#   Convert Rig
#-------------------------------------------------------------

SourceRig = {
    "genesis" : "genesis1",
    "genesis_2_female" : "genesis2",
    "genesis_2_male" : "genesis2",
    "genesis_3_female" : "genesis3",
    "genesis_3_male" : "genesis3",
    "genesis_8_female" : "genesis8",
    "genesis_8_male" : "genesis8",
}

class DAZ_OT_ConvertRigPose(DazPropsOperator, B.NewRig):
    bl_idname = "daz.convert_rig"
    bl_label = "Convert DAZ Rig"
    bl_description = "Convert current DAZ rig to other DAZ rig"
    bl_options = {'UNDO'}

    @classmethod
    def poll(self, context):
        ob = context.object
        return (ob and ob.type == 'ARMATURE' and ob.DazRig[0:7] == "genesis")

    def draw(self, context):
        self.layout.prop(self, "newRig")

    def run(self, context):
        from .globvars import theRestPoseFolder
        global RestPoses

        rig = context.object
        scn = context.scene
        loadRestPoseEntry(self.newRig, RestPoses, theRestPoseFolder)
        scale = 1.0
        if self.newRig in SourceRig.keys():
            modify = False
            src = SourceRig[self.newRig]
            conv,twists = getConverter(src, rig)
            if conv:
                self.renameBones(rig, conv)
        else:
            modify = True
            src = self.newRig
            table = RestPoses[src]
            if "translate" in table.keys():
                self.renameBones(rig, table["translate"])
            if "scale" in table.keys():
                scale = table["scale"] * rig.DazScale
        loadPose(rig, self.newRig, RestPoses, modify)
        rig.DazRig = src
        print("Rig converted to %s" % self.newRig)
        if scale != 1.0:
            raise DazError("Use scale = %.5f when loading BVH files.       " % scale, True)


    def renameBones(self, rig, conv):
        bpy.ops.object.mode_set(mode='EDIT')
        for eb in rig.data.edit_bones:
            if eb.name in conv.keys():
                data = conv[eb.name]
                if isinstance(data, list):
                    eb.name = data[0]
                    if data[1] == "reverse":
                        head = tuple(eb.head)
                        tail = tuple(eb.tail)
                        eb.head = (1,2,3)
                        eb.tail = head
                        eb.head = tail
                else:
                    eb.name = data
        bpy.ops.object.mode_set(mode='OBJECT')

#-------------------------------------------------------------
#   Bone conversion
#-------------------------------------------------------------

TwistBones["genesis3"] = [
    ("lShldrBend", "lShldrTwist"),
    ("rShldrBend", "rShldrTwist"),
    ("lForearmBend", "lForearmTwist"),
    ("rForearmBend", "rForearmTwist"),
    ("lThighBend", "lThighTwist"),
    ("rThighBend", "rThighTwist"),
]
TwistBones["genesis8"] = TwistBones["genesis3"]


def getConverter(stype, trg):
    if stype == "genesis8":
        stype = "genesis3"
    trgtype = trg.DazRig
    if trgtype == "genesis8":
        trgtype = "genesis3"

    if stype == "" or trgtype == "":
        return {},[]
    if (stype in TwistBones.keys() and
        trgtype not in TwistBones.keys()):
        twists = TwistBones[stype]
    else:
        twists = []

    if stype == trgtype:
        return {},twists
    if trgtype == "mhx":
        cname = stype[:-1] + "-mhx"
    elif trgtype[0:6] == "rigify":
        cname = stype[:-1] + "-" + trgtype
    else:
        cname = stype + "-" + trgtype

    conv = getConverterEntry(cname)
    if not conv:
        print("No converter", stype, trg.DazRig)
    return conv, twists


def getConverterEntry(cname):
    import json
    from .fileutils import safeOpen
    if cname in Converters.keys():
        return Converters[cname]
    else:
        folder = os.path.join(os.path.dirname(__file__), "data", "converters")
        filepath = os.path.join(folder, cname + ".json")
        if os.path.exists(filepath):
            with safeOpen(filepath, "r") as fp:
                conv = Converters[cname] = json.load(fp)
            return conv
    return {}

#----------------------------------------------------------
#   Initialize
#----------------------------------------------------------

classes = [
    DAZ_OT_SaveCurrentPose,
    DAZ_OT_LoadPose,
    DAZ_OT_OptimizePose,
    DAZ_OT_ConvertRigPose,
]

def initialize():
    for cls in classes:
        bpy.utils.register_class(cls)


def uninitialize():
    for cls in classes:
        bpy.utils.unregister_class(cls)
